// Copyright (c) 2024 FWORKS. All Rights Reserved.

#include "SuperSimpleFSMComponent.h"
#include "GameplayTagsManager.h"
#include "Net/UnrealNetwork.h"
#include "GameFramework/Actor.h"
#include "UnrealEngine.h"

TAutoConsoleVariable<int32> CVarLogSSFSMStateTransitions(
    TEXT("SuperSimpleFSM.LogStateTransitions"),
    0,
    TEXT("Log state enters and exits"),
    ECVF_Default);

DEFINE_LOG_CATEGORY(LogSuperSimpleFSM);

// Sets default values for this component's properties
USuperSimpleFSMComponent::USuperSimpleFSMComponent()
{
	PrimaryComponentTick.bCanEverTick = true;
	PrimaryComponentTick.bStartWithTickEnabled = false;
}

void USuperSimpleFSMComponent::GetLifetimeReplicatedProps(TArray<FLifetimeProperty>& OutLifetimeProps) const
{
	Super::GetLifetimeReplicatedProps(OutLifetimeProps);

	DOREPLIFETIME(USuperSimpleFSMComponent, ActiveStateTag);
}

FGameplayTag USuperSimpleFSMComponent::GetActiveStateTag() const
{
	return ActiveStateTag;
}

bool USuperSimpleFSMComponent::IsStateActive(FGameplayTag CheckTag) const
{
	return ActiveStateTag.MatchesTag(CheckTag);
}

bool USuperSimpleFSMComponent::CanPerformTransition(FGameplayTag NewActiveStateTag) const
{
	if (!bTransitionAllowed)
	{
		return false;
	}

	if (!GetOwner())
	{
		return false;
	}

	if (GetIsReplicated() && !GetOwner()->HasAuthority())
	{
		return false;
	}

	return true;
}

void USuperSimpleFSMComponent::PerformTransition(FGameplayTag NewActiveStateTag)
{
	if (!CanPerformTransition(NewActiveStateTag))
	{
		return;
	}

	InternalHandleTransition(NewActiveStateTag, ActiveStateTag);
}

void USuperSimpleFSMComponent::OnRegisterStateCallbacks_Implementation()
{
	// for child c++ classes to register callbacks
}

void USuperSimpleFSMComponent::OnRep_ActiveStateTag(FGameplayTag OldStateTag)
{
	// revert it back temporarily, let the HandleTransition change the tag
	const FGameplayTag NewTag = ActiveStateTag;
	ActiveStateTag = OldStateTag;
	InternalHandleTransition(NewTag, OldStateTag);
}

void USuperSimpleFSMComponent::RefreshNeedsToTick()
{
	if (FSMNeedsTick())
	{
		SetComponentTickEnabled(true);
	}
	else
	{
		SetComponentTickEnabled(false);
	}
}

bool USuperSimpleFSMComponent::FSMNeedsTick() const
{
	return TickingStates.Num() > 0;
}

void USuperSimpleFSMComponent::BeginPlay()
{
	Super::BeginPlay();

	bFSMCallbackRegistrationAllowed = true;
	OnRegisterStateCallbacks();
	bFSMCallbackRegistrationAllowed = false;

	// at this point, all our state callbacks will be ready
	const bool bGoToInitialState = !GetIsReplicated() || (GetOwner() && GetOwner()->HasAuthority());
	if (bGoToInitialState)
	{
		if (InitialState.IsValid())
		{
			InternalHandleTransition(InitialState, FGameplayTag::EmptyTag);
		}
	}
	else
	{
		if (ActiveStateTag.IsValid())
		{
			InternalHandleTransition(ActiveStateTag, FGameplayTag::EmptyTag);
		}
	}
}

void USuperSimpleFSMComponent::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	// exit all our states

	if (ActiveStateTag.IsValid())
	{
		InternalHandleTransition(FGameplayTag::EmptyTag, ActiveStateTag);
	}

	Super::EndPlay(EndPlayReason);
}

FGameplayTag USuperSimpleFSMComponent::GetActiveChild(FGameplayTag Parent) const
{
	// eg ActiveStateTag is A.B.C.D.E.F
	// Parent is A.B
	// we want to return A.B.C

	if (!IsStateActive(Parent))
	{
		return FGameplayTag::EmptyTag;
	}

	TArray<FGameplayTag> ParentTagArray;
	Parent.GetGameplayTagParents().GetGameplayTagArray(ParentTagArray);
	TArray<FGameplayTag> ActiveStateTagArray;
	ActiveStateTag.GetGameplayTagParents().GetGameplayTagArray(ActiveStateTagArray);
	const int32 ParentCount = ParentTagArray.Num();
	const int32 ActiveCount = ActiveStateTagArray.Num();
	const int32 DesiredIndex = ActiveCount - 1 - ParentCount;

	if (DesiredIndex < 0 || ParentCount <= 0)
	{
		return FGameplayTag::EmptyTag;
	}

	return ActiveStateTagArray[DesiredIndex];
}

void USuperSimpleFSMComponent::TickComponent(float DeltaTime, enum ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	// component tick is off by default, until someone registers a tick onto a state and that state becomes active
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	TArray<FGameplayTag> TagArray;
	ActiveStateTag.GetGameplayTagParents().GetGameplayTagArray(TagArray);
	if (TagArray.Num() > 0)
	{
		FGameplayTag ActiveChild = TagArray.Last(); // the root state
		while (ActiveChild != FGameplayTag::EmptyTag)
		{
			FSuperSimpleFSMStateTickDelegate* ExistingTick = StateTickCallbacks.Find(ActiveChild);
			if (ExistingTick)
			{
				ExistingTick->ExecuteIfBound(ActiveChild, DeltaTime);
			}

			//GEngine->AddOnScreenDebugMessage(-1, 0.0f, FColor::Yellow, FString::Printf(TEXT("Ticking %s"), *ActiveChild.ToString()));
			ActiveChild = GetActiveChild(ActiveChild);
		}

	}
}

void USuperSimpleFSMComponent::RegisterStateEvents(FGameplayTag StateTag, FSuperSimpleFSMStateDelegate OnBegin, FSuperSimpleFSMStateDelegate OnEnd, FSuperSimpleFSMStateTickDelegate OnUpdate)
{
	if (!bFSMCallbackRegistrationAllowed)
	{
		UE_LOG(LogSuperSimpleFSM, Log, TEXT("Actor(%s) trying to register state events for %s outside of OnRegisterStateCallbacks, it will be ignored."), *GetNameSafe(GetOwner()), *StateTag.ToString());
		return;
	}

	if (OnBegin.IsBound())
	{
		FSuperSimpleFSMStateDelegate* ExistingBegin = StateBeginCallbacks.Find(StateTag);
		if (ExistingBegin)
		{
			StateBeginCallbacks.Remove(StateTag);
		}

		StateBeginCallbacks.Add(StateTag, OnBegin);
	}

	if (OnEnd.IsBound())
	{
		FSuperSimpleFSMStateDelegate* ExistingBegin = StateEndCallbacks.Find(StateTag);
		if (ExistingBegin)
		{
			StateEndCallbacks.Remove(StateTag);
		}

		StateEndCallbacks.Add(StateTag, OnEnd);
	}

	if (OnUpdate.IsBound())
	{
		FSuperSimpleFSMStateTickDelegate* ExistingTick = StateTickCallbacks.Find(StateTag);
		if (ExistingTick)
		{
			StateTickCallbacks.Remove(StateTag);
		}

		StateTickCallbacks.Add(StateTag, OnUpdate);
	}
}

void USuperSimpleFSMComponent::InternalOnEnter(FGameplayTag EnterTag)
{
    if (CVarLogSSFSMStateTransitions.GetValueOnGameThread() > 0)
    {
        UE_LOG(LogSuperSimpleFSM, Log, TEXT("Enter %s (%s)"), *EnterTag.ToString(), *GetDebugStringForWorld(GetWorld()));
    }

    OnStateBegin.Broadcast(EnterTag);

    FSuperSimpleFSMStateTickDelegate* ExistingTick = StateTickCallbacks.Find(EnterTag);
    if (ExistingTick)
    {
        TickingStates.AddTag(EnterTag);
    }

	FSuperSimpleFSMStateDelegate* BeginCallback = StateBeginCallbacks.Find(EnterTag);
	if (BeginCallback)
	{
		BeginCallback->ExecuteIfBound(EnterTag);
	}
}

void USuperSimpleFSMComponent::InternalOnExit(FGameplayTag ExitTag)
{
	bTransitionAllowed = false;

    if (CVarLogSSFSMStateTransitions.GetValueOnGameThread() > 0)
    {
        UE_LOG(LogSuperSimpleFSM, Log, TEXT("Exit %s (%s)"), *ExitTag.ToString(), *GetDebugStringForWorld(GetWorld()));
    }

	FSuperSimpleFSMStateDelegate* EndCallback = StateEndCallbacks.Find(ExitTag);
	if (EndCallback)
	{
		EndCallback->ExecuteIfBound(ExitTag);
	}

	bTransitionAllowed = true;

	OnStateEnd.Broadcast(ExitTag);

	FSuperSimpleFSMStateTickDelegate* ExistingTick = StateTickCallbacks.Find(ExitTag);
	if (ExistingTick)
	{
		TickingStates.RemoveTag(ExitTag);
	}

	//UE_LOG(LogTemp, Log, TEXT("Exit %s (%s) : state is now: %s"), *OldSubTag.ToString(), *GetDebugStringForWorld(GetWorld()), *ActiveStateTag.ToString());
}

void USuperSimpleFSMComponent::InternalHandleTransition(FGameplayTag NewTag, FGameplayTag OldTag)
{
	if (NewTag == OldTag)
	{
		return;
	}

	// should we only allow transitioning to leaf tags?
	//const FGameplayTagContainer ChildTags = UGameplayTagsManager::Get().RequestGameplayTagChildren(NewTag);
	//if (!ChildTags.IsEmpty())
	//{
	//	UE_LOG(LogSuperSimpleFSM, Log, TEXT("Actor(%s) trying to transition to a non-leaf state (%s) it will be ignored."),*GetNameSafe(GetOwner()), *StateTag.ToString());
	//	return;
	//}

	TArray<FGameplayTag> NewSubTags;
	NewTag.GetGameplayTagParents().GetGameplayTagArray(NewSubTags);
	if (!OldTag.IsValid())
	{
		// just enter all the new states, we need to go in reverse
		for (int32 i = NewSubTags.Num() - 1; i >= 0; i--)
		{
			ActiveStateTag = NewSubTags[i];
			InternalOnEnter(NewSubTags[i]);

			if (ActiveStateTag != NewSubTags[i])
			{
				// a transition was performed in that state's begin, we shall abort
				return;
			}
		}

		RefreshNeedsToTick();
		return;
	}

	TArray<FGameplayTag> ActiveSubTags;
	OldTag.GetGameplayTagParents().GetGameplayTagArray(ActiveSubTags);
	if (!NewTag.IsValid())
	{
		// just exit all the existing states, go in forward
		for (int32 i = 0; i < ActiveSubTags.Num(); i++)
		{
			ActiveStateTag = (i < ActiveSubTags.Num() - 1) ? ActiveSubTags[i + 1] : FGameplayTag::EmptyTag;
			InternalOnExit(ActiveSubTags[i]);
		}

		RefreshNeedsToTick();
		return;
	}

	// both old and new state is valid
	// eg A.B.C -> A.B.E.F
	// we want to (order is important) 
	// - exit C
	// - enter E
	// - enter F

	FGameplayTag TempNewTag = FGameplayTag::EmptyTag;

	// go up the old tags until we find a tag that's present in the new tag
	for (int32 i = 0; i < ActiveSubTags.Num(); i++)
	{
		const FGameplayTag Curr = ActiveSubTags[i];
		if (!NewTag.MatchesTag(Curr))
		{
			ActiveStateTag = (i < ActiveSubTags.Num() - 1) ? ActiveSubTags[i + 1] : FGameplayTag::EmptyTag;
			InternalOnExit(Curr);
		}
		else
		{
			TempNewTag = Curr;
			break;
		}
	}

	for (int32 i = NewSubTags.Num() - 1; i >= 0; i--)
	{
		const FGameplayTag Curr = NewSubTags[i];
		if (TempNewTag.MatchesTag(Curr))
		{
			// keep going
		}
		else
		{
			ActiveStateTag = Curr;
			InternalOnEnter(Curr);

			if (ActiveStateTag != Curr)
			{
				// a transition was performed in that state's begin, we shall abort
				RefreshNeedsToTick();
				return;
			}
		}
	}

	RefreshNeedsToTick();
}
