// Copyright (c) 2024 FWORKS. All Rights Reserved.

#pragma once

#include "CoreMinimal.h"
#include "Components/ActorComponent.h"
#include "GameplayTagContainer.h"
#include "Delegates/Delegate.h"
#include "SuperSimpleFSMComponent.generated.h"

DECLARE_LOG_CATEGORY_EXTERN(LogSuperSimpleFSM, Log, All);
DECLARE_DYNAMIC_DELEGATE_OneParam(FSuperSimpleFSMStateDelegate, FGameplayTag, StateTag);
DECLARE_DYNAMIC_DELEGATE_TwoParams(FSuperSimpleFSMStateTickDelegate, FGameplayTag, StateTag, float, DeltaTime);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FSuperSimpleFSMStateEnterExitDelegate, FGameplayTag, StateTag);

/*
 * Simple FSM functionality that uses gameplay tags to support hierarchical state management
 * - Override OnRegisterStateCallbacks in your BP/C++ class
 * - Call RegisterStateEvents within that function to point to events in your BP/C++
 * - Set the InitialState on the component
 * - Call PerformTransition whenever you want to change state
 * - (Optional) Turn on component replicates if you want transitions to be done on server and active state replicated to client
 *		Note this replicates only the state, and not transitions
 */
UCLASS( ClassGroup=(Custom), meta=(BlueprintSpawnableComponent), Blueprintable )
class SUPERSIMPLEFSM_API USuperSimpleFSMComponent : public UActorComponent
{
	GENERATED_BODY()

public:	
	// Sets default values for this component's properties
	USuperSimpleFSMComponent();

	virtual void BeginPlay() override;
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	virtual void TickComponent(float DeltaTime, enum ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;
	virtual void GetLifetimeReplicatedProps(TArray<FLifetimeProperty>& OutLifetimeProps) const override;

	// checks whether a given state is active (including all its child tags)
	UFUNCTION(BlueprintCallable, Category="SimpleFSM")
	bool IsStateActive(FGameplayTag CheckTag) const;

	// returns the current active state hierarchy
	UFUNCTION(BlueprintCallable, Category="SuperSimpleFSM")
	FGameplayTag GetActiveStateTag() const;

	// performs a transition to another state. If the component is replicated, this call will be ignored on non-server
	UFUNCTION(BlueprintCallable, Category="SuperSimpleFSM")
	void PerformTransition(FGameplayTag NewActiveStateTag);

	// Callback for whenever a state begins
	UPROPERTY(Transient, BlueprintAssignable, Category = "SuperSimpleFSM Debug")
	FSuperSimpleFSMStateEnterExitDelegate OnStateBegin;

	// Callback for whenever a state ends
	UPROPERTY(Transient, BlueprintAssignable, Category = "SuperSimpleFSM Debug")
	FSuperSimpleFSMStateEnterExitDelegate OnStateEnd;
	
private:
	void InternalHandleTransition(FGameplayTag NewTag, FGameplayTag OldTag);
	void InternalOnEnter(FGameplayTag EnterTag);
	void InternalOnExit(FGameplayTag ExitTag);
	void RefreshNeedsToTick();
	FGameplayTag GetActiveChild(FGameplayTag Parent) const;

protected:

	// Register your state events here
	UFUNCTION(BlueprintNativeEvent, Category = "SuperSimpleFSM")
	void OnRegisterStateCallbacks();

	// Register Begin/End/Update events. Do not bind OnUpdate if you don't need it.
	UFUNCTION(BlueprintCallable, Category = "SuperSimpleFSM")
	void RegisterStateEvents(FGameplayTag StateTag, FSuperSimpleFSMStateDelegate OnBegin, FSuperSimpleFSMStateDelegate OnEnd, FSuperSimpleFSMStateTickDelegate OnUpdate);

	UFUNCTION()
	void OnRep_ActiveStateTag(FGameplayTag OldStateTag);

	virtual bool CanPerformTransition(FGameplayTag NewActiveStateTag) const;
	virtual bool FSMNeedsTick() const;

	// initial starting state
	UPROPERTY(EditAnywhere, Category = "SuperSimpleFSM")
	FGameplayTag InitialState = FGameplayTag::EmptyTag;

	// the current active state hierarchy
	UPROPERTY(Transient, Category = "SuperSimpleFSM Debug", VisibleAnywhere, ReplicatedUsing = OnRep_ActiveStateTag)
	FGameplayTag ActiveStateTag;

	// all states that are actively ticking
	UPROPERTY(Transient, Category = "SuperSimpleFSM Debug", VisibleAnywhere)
	FGameplayTagContainer TickingStates;
	
	// all registered begin events
	UPROPERTY(Transient, Category = "SuperSimpleFSM Debug", VisibleAnywhere)
	TMap<FGameplayTag, FSuperSimpleFSMStateDelegate> StateBeginCallbacks;

	// all registered end events
	UPROPERTY(Transient, Category = "SuperSimpleFSM Debug", VisibleAnywhere)
	TMap<FGameplayTag, FSuperSimpleFSMStateDelegate> StateEndCallbacks;

	// all registered tick events
	UPROPERTY(Transient, Category = "SuperSimpleFSM Debug", VisibleAnywhere)
	TMap<FGameplayTag, FSuperSimpleFSMStateTickDelegate> StateTickCallbacks;

	bool bFSMCallbackRegistrationAllowed = false;
	bool bTransitionAllowed = true;
};
